/**********************************************************************
 * pop3_in.c                                             September 1999
 * Horms                                             horms@vergenet.net
 *
 * Handle pop commands from a client
 *
 * Licenced under the GNU GPL
 *
 **********************************************************************/

#include "pop3_in.h"


/**********************************************************************
 * pop3_in_authenticate
 * Authenticate an incoming pop session
 * Not really needed if we are going to authenticate with an upstream
 * pop server but it may be useful in some cases
 *
 * Return 1 on success, 0 on failure and -1 on error.
 **********************************************************************/

int pop3_in_authenticate(
  const struct passwd *pw, 
  const int err_fd,
  const token_t *tag
){
  pam_handle_t *pamh=NULL;

  extern int pam_retval;
  extern struct pam_conv conv_struct;

  if((
     pam_retval=pam_start(SERVICE_NAME, pw->pw_name, &conv_struct, &pamh)
  )!=PAM_SUCCESS){
    syslog(LOG_WARNING, "main: pam_start: %s", strerror(errno));
    do_pam_end(pamh, EX_PAM_ERROR);
    return(-1);
  }
  if(do_pam_authentication(pamh, pw->pw_name, pw->pw_passwd)<0){
    sleep(PERDITION_AUTH_FAIL_SLEEP);
    if(pop3_write(err_fd, NULL, POP3_ERR,"Authentication failure")<0){
      syslog(LOG_WARNING, "main: pop3_write");
      do_pam_end(pamh, EXIT_SUCCESS);
      return(-1);
    }
    do_pam_end(pamh, EXIT_SUCCESS);
    return(0);
  }

  do_pam_end(pamh, EXIT_SUCCESS);
  return(1);
}


/**********************************************************************
 * pop3_in_get_pw
 * read USER and PASS commands and return them in a struct passwd *
 * allocated by this function
 * Pre: in_fd: file descriptor to read from
 *      out_fd: file descriptor to write to
 *      return_tag:   ignored 
 * Post: passwd structure with pw_name and pw_passwd set,
 *       NULL on error
 **********************************************************************/

struct passwd *pop3_in_get_pw(
  const int in_fd, 
  const int out_fd,
  token_t ** return_tag
){
  queue *q = NULL;
  struct passwd *pw;
  char *command_string;
  token_t *t = NULL;
  char *message=NULL;

  command_string=NULL;

  if((pw=(struct passwd *)malloc(sizeof(struct passwd)))==NULL){
    syslog(LOG_WARNING, "pop3_in_get_pw: malloc: %s", strerror(errno));
    destroy_queue(q, DESTROY_TOKEN);
    return(NULL);
  }

  pw->pw_name=NULL;

  while(1){
    if((q=read_line(in_fd, q))==NULL){
      syslog(LOG_WARNING, "pop3_in_get_pw: read_pop3_line");
      break;
    }

    if((q=pop(q, (queue_t *)&t))==NULL){
      syslog(LOG_WARNING, "pop3_in_get_pw: pop");
      t=NULL;
      break;
    }

    if((command_string=token_to_string(t, command_string))==NULL){
      syslog(LOG_WARNING, "pop3_in_get_pw: token_to_string");
      break;
    }
    unassign_token(t);
    destroy_token(&t);

    if(strcasecmp(command_string, "USER")==0){
      if(pw->pw_name!=NULL){
        sleep(PERDITION_ERR_SLEEP);
        if(pop3_write(out_fd, NULL, POP3_ERR, "USER is already set")<0){
          syslog(LOG_WARNING, "pop3_in_get_pw: pop3_write 1");
          break;
        }
        goto loop;
      }
      if(queue_length(q)!=1){
        sleep(PERDITION_ERR_SLEEP);
        if(pop3_write(out_fd, NULL, POP3_ERR, "Try USER <username>")<0){
          syslog(LOG_WARNING, "pop3_in_get_pw: pop3_write 2");
          break;
        }
        goto loop;
      }

      if((q=pop(q, (queue_t *)&t))==NULL){
        syslog(LOG_WARNING, "pop3_in_get_pw: pop");
        t=NULL;
        break;
      }
      if((pw->pw_name=token_to_string(t, pw->pw_name))==NULL){
        syslog(LOG_WARNING, "pop3_in_get_pw: token_to_string");
        break;
      }
      unassign_token(t);
      destroy_token(&t);
      
      if((message=cat_str(message, 3, "USER ", pw->pw_name, " set"))<0){
        syslog(LOG_WARNING, "pop3_in_get_pw: cat_str");
        goto loop;
      }
      if(pop3_write(out_fd, NULL, POP3_OK, message)<0){
        syslog(LOG_WARNING, "pop3_in_get_pw: pop3_write");
        goto loop;
      }
    }
    else if(strcasecmp(command_string, "PASS")==0){
      str_free(&command_string);
      if(pw->pw_name==NULL){
        sleep(PERDITION_ERR_SLEEP);
        if(pop3_write(out_fd, NULL, POP3_ERR, "USER not yet set")<0){
          syslog(LOG_WARNING, "pop3_in_get_pw: pop3_write");
          break;
        }
        goto loop;
      }
      if(queue_length(q)!=1){
        sleep(PERDITION_ERR_SLEEP);
        if(pop3_write(out_fd, NULL, POP3_ERR, "Try PASS <password>")<0){
          syslog(LOG_WARNING, "pop3_in_get_pw: pop3_write");
          break;
        }
        goto loop;
      }
      if((q=pop(q, (queue_t *)&t))==NULL){
        syslog(LOG_WARNING, "pop3_in_get_pw: pop");
        free(pw->pw_name);
        break;
      }
      if((pw->pw_passwd=token_to_string(t, pw->pw_passwd))==NULL){
        syslog(LOG_WARNING, "pop3_in_get_pw: token_to_string");
        free(pw->pw_name);
        break;
      }
      unassign_token(t);
      destroy_token(&t);
      destroy_queue(q, DESTROY_TOKEN);
      return(pw);
    }
    else if(strcasecmp(command_string, "QUIT")==0){
      syslog(LOG_INFO, "Closing NULL session: username=%s", pw->pw_name);
      if(pop3_write(out_fd, NULL, POP3_OK, "QUIT")<0){
        syslog(LOG_WARNING, "pop3_in_get_pw: pop3_write");
        break;
      }
      destroy_queue(q, DESTROY_TOKEN);
      daemon_exit_cleanly(0);
    }
    else{
      sleep(PERDITION_ERR_SLEEP);
      if(pop3_write(
        out_fd, 
	NULL,
        POP3_ERR,
        " Command must be one of USER, PASS or QUIT"
      )<0){
        syslog(LOG_WARNING, "pop3_in_get_pw: pop3_write");
        break;
      }
    }

    /*Clean up before looping*/
    loop:
    destroy_token(&t);
    str_free(&command_string);
    destroy_queue(q, DESTROY_TOKEN);
    str_free(&message);
  }

  /*If we get here clean up and bail*/
  destroy_token(&t);
  str_free(&command_string);
  destroy_queue(q, DESTROY_TOKEN);
  free(pw);
  str_free(&message);
  return(NULL);
}


/**********************************************************************
 * pop3_in_greeting
 * Send a greeting to the user
 * Pre: fd: file descriptor to write to
 *      tag: ignored
 * Return 1 on success
 *        0 otherwise
 **********************************************************************/

int pop3_in_greeting(const int fd){
  char *message=NULL;

  extern struct utsname *system_uname;

  if((message=cat_str(
    message, 
    3, 
    POP3_GREETING, 
    " ", 
    system_uname->nodename
  ))==NULL){
    syslog(LOG_WARNING, "pop3_in_greeting: cat_str");
    return(-1);
  }

  if(pop3_write(fd, NULL, POP3_OK, message)<0){
    syslog(LOG_WARNING, "pop3_in_greeting: pop3_write");
    return(-1);
  }

  free(message);

  return(0);
}
