/**********************************************************************
 * pop3.c                                                September 1999
 * Horms                                             horms@vergenet.net
 *
 * Functions to communicate with upstream POP3 server
 **********************************************************************/

#include "pop3_out.h"

/**********************************************************************
 * pop3_out_authenticate
 * Authenticate user with backend pop3 server
 * pre: in_fd: File descriptor to read server responses from
 *      out_fd: File descriptor to write to server on
 *      err_fd: File descriptor to write error to
 *      tag:    ignored 
 *      pw:     structure with username and passwd
 * post: 1: on success
 *       0: on failure
 *       -1 on error
 **********************************************************************/

int pop3_out_authenticate(
  const int in_fd, 
  const int out_fd, 
  const struct passwd *pw,
  protocol_t *protocol
){
  token_t *ok;
  int status;
  char *read_string;
  char *greeting_string;
  queue *q;

  read_string=NULL;
  greeting_string=NULL;
  status=-1;

  if((ok=create_token(ok))==NULL){
    syslog(LOG_WARNING, "pop3_out_get_pw: create_token");
    goto leave;
  }
  assign_token(ok, (PERDITION_USTRING) POP3_OK, strlen(POP3_OK), -1);

  if((status=pop3_out_response(in_fd, NULL, ok, &q))<0){
    syslog(LOG_WARNING, "pop3_out_authenticate: pop3_out_response 1");
    goto leave;
  }
  else if(!status){
    status=0;
    goto leave;
  }

  if((read_string=queue_to_string(q, read_string))==NULL){
    syslog(LOG_WARNING, "pop3_out_authenticate: queue_to_string");
    status=-1;
    goto leave;
  }

  if((greeting_string=greeting_str(greeting_string, protocol))==NULL){
    syslog(LOG_WARNING, "pop3_out_authenticate: greeting_str");
    status=-1;
    goto leave;
  }

  if((status=strcmp(read_string, greeting_string))==0){
    syslog(LOG_WARNING, "Loop detected, abandoning connection");
    goto leave;
  }

  if(pop3_write(out_fd, NULL, "USER", pw->pw_name)<0){
    syslog(LOG_WARNING, "pop3_out_get_pw: pop3_write");
    goto leave;
  }

  if((status=pop3_out_response(in_fd, NULL, ok, &q))<0){
    syslog(LOG_WARNING, "pop3_out_authenticate: pop3_out_response 2");
    goto leave;
  }
  else if(!status){
    status=0;
    goto leave;
  }

  destroy_queue(q, DESTROY_TOKEN);

  if(pop3_write(out_fd, NULL, "PASS", pw->pw_passwd)<0){
    syslog(LOG_WARNING, "pop3_out_get_pw: pop3_write");
    return(-1);
  }

  if((status=pop3_out_response(in_fd, NULL, ok, &q))<0){
    syslog(LOG_WARNING, "pop3_out_authenticate: pop3_out_response 3");
  }
  destroy_queue(q, DESTROY_TOKEN);

  leave:
  str_free(&read_string);
  str_free(&greeting_string);
  destroy_token(&ok);
  return(status);
}
  

/**********************************************************************
 * pop3_out_response
 * Compare a respnse from a server with the desired response
 * pre: in_fd: file descriptor to read from
 *      out_fd: file descriptor to write to
 *      tag_string: ignored
 *      desired_token: token expected from server
 * post: 1 : tag and desired token found
 *       0: tag and desired token not found
 *       -1: on error
 **********************************************************************/


int pop3_out_response(
  int in_fd, 
  const char *tag_string,
  const token_t *desired_token,
  queue **q
){
  int status;
  token_t *t;

  if((*q=read_line(in_fd, *q))==NULL){
    syslog(LOG_WARNING, "pop3_out_response: read_line");
    return(-1);
  }
  if((*q=pop(*q, (queue_t *)&t))==NULL){
    syslog(LOG_WARNING, "pop3_out_response: pop");
    return(-1);
  }

  status=token_cmp(desired_token, t);
  
  destroy_token(&t);
  return(status);
}
